/***************************************************************************
 *                                                                         *
 *   KNetLoad is copyright (c) 1999-2000, Markus Gustavsson                *
 *                         (c) 2002, Ben Burton                            *
 *                         (c) 2004, Diego Pettenò                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "knetproc.h"

#include <kiconloader.h>
#include <kglobal.h>
#include <qpixmap.h>

// BSD-specific includes.
#ifdef Q_OS_BSD4
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <net/if_dl.h>
#include <net/route.h>
#include <net/if.h>
#endif

// Solaris-specific includes.
#ifdef Q_OS_SOLARIS
#include <kstat.h>
#include <sys/sockio.h>
#endif

// Linux-specific includes.
#ifdef Q_OS_LINUX
#include <qdir.h>
#include <qregexp.h>
#include <qfile.h>
#endif

QStringList KNetProc::s_interfaces;

/**
 * Linux
 * -----
 *
 * For kernels 2.4 and earlier, we use /proc/net/dev file, which contains
 * lines structured in this way:
 * 
 *   iface: bytesIn x x x x x x x bytesOut x x x x x x x
 * 
 * The first two lines are headers, but also if they are not, there are
 * no problems: the regular expression make simpler to parse it.
 * 
 * For kernels 2.5 and greater, if /sys directory is present, it's used
 * to get the statistics.
 *
 * BSD
 * ---
 *
 * BSD code by Roland Riegel <feedback@roland-riegel.de>,
 * Bill Wilson <bill@gkrellm.net>, Hajimu Umemoto <ume@mahoroba.org> and
 * Anthony Mallet <anthony.mallet@useless-ficus.net>.
 *
 * Solaris
 * -------
 *
 * Solaris code by Roland Riegel <feedback@roland-riegel.de>,
 * Bill Wilson <bill@gkrellm.net> and Daisuke Yabuki <dxy@acm.org>.
 *
 * Other Operating Systems
 * -----------------------
 *
 * Please, send in a patch!
 *
 * The KNetLoad maintainer is currently Diego Pettenò
 * <flameeyes@users.berlios.de>, or you could submit a patch through the KDE
 * bug tracking system ( http://bugs.kde.org/ ).
 */

#define MAX_NET_DEV_LINE 512
#define MAX_INTERFACE_NAME 128

const float KNetProc::megabyte = 1000000;

KNetProc::KNetProc() {
    currIn = currOut = 0;
    prevIn = prevOut = 0;

    // OS-specific initialisation.

    // ========== BSD-specific (begin) ==========
#ifdef Q_OS_BSD4
    buf = 0;
    alloc = 0;
#endif
    // ========== BSD-specific (end) ==========
    
    // ========== Linux-specific (begin) ==========
#ifdef Q_OS_LINUX
    useSys = QDir::home().exists("/sys/class/net");
#endif
    // ========== Linux-specific (end) ==========
}

KNetProc::~KNetProc() {
    // OS-specific cleaning up.

    // ========== BSD-specific (begin) ==========
#ifdef Q_OS_BSD4
    if (buf)
        free(buf);
#endif
    // ========== BSD-specific (end) ==========
}

void KNetProc::setDevice(const QString &interface) {
    // Actually set the device.
    device = interface;

    // Take an initial reading.
    currIn = currOut = 0;
    prevIn = prevOut = 0;
    readLoad();

    // Make the previous reading identical to the current reading to
    // avoid a massive spike before the next reading is taken.
    prevIn = currIn;
    prevOut = currOut;
}

void KNetProc::readLoad() {
    // The current readings must now become the previous readings.
    prevIn = currIn;
    prevOut = currOut;

    // Take the new set of readings (i.e., update currIn and currOut).

    // If no new readings are taken (e.g., the device could not be
    // found), this code may return; otherwise it *must* fall through to
    // the tidy-up block that follows this OS-specific code.

#ifdef Q_OS_LINUX
    if ( useSys )
        readLoadLinux26();
    else
        readLoadLinux24();
#elif defined Q_OS_BSD4
    readLoadBSD4();
#elif defined Q_OS_SOLARIS
    readLoadSolaris();
#endif

    // Tidy up.
    if (currIn < prevIn)
        prevIn = 0;
    if (currOut < prevOut)
        prevIn = 0;
}

// Linux specific function
void KNetProc::readLoadLinux26() {
#ifdef Q_OS_LINUX
	static QFile rx("/sys/class/net/" + device + "/statistics/rx_bytes");
	static QFile tx("/sys/class/net/" + device + "/statistics/tx_bytes");
	QString read;
	
	if ( ! rx.open(IO_ReadOnly) )
		return;
	
	if ( rx.readLine(read, 32) == -1 )
	{
		rx.close();
		return;
	}
	
	read.stripWhiteSpace();
	currIn = read.toUInt();
	
	rx.close();
	
	if ( ! tx.open(IO_ReadOnly) )
		return;
	
	if ( tx.readLine(read, 32) == -1 )
	{
		tx.close();
		return;
	}
	
	read.stripWhiteSpace();
	currOut = read.toUInt();
	
	tx.close();
#endif
}

void KNetProc::readLoadLinux24() {
#ifdef Q_OS_LINUX
	static QFile proc("/proc/net/dev");
	static QRegExp procre("^\\s+(\\S+):(\\d+)\\s+(\\d+\\s+){7}(\\d+\\s+)(\\d+\\s*){7}$");
	QString read;
	
	if ( ! proc.open(IO_ReadOnly) )
		return;
	
	while ( ! proc.atEnd() )
	{
		if ( proc.readLine(read, MAX_NET_DEV_LINE) == -1 )
		{
			proc.close();
			return;
		}
		
		if ( procre.search(read) == -1 )
			continue;
		
		if ( procre.cap(1) != device )
			continue;
		
		currIn = procre.cap(2).toUInt();
		currOut = procre.cap(4).toUInt();
		break;
	}
	
	proc.close();
#endif
}

// Solaris specific
void KNetProc::readLoadBSD4() {
#ifdef Q_OS_BSD4
    static struct if_msghdr *ifm, *nextifm;
    static struct sockaddr_dl *sdl;
    static char *lim, *next;
    static size_t needed;
    static char s[32];
    static int mib[] = { CTL_NET, PF_ROUTE, 0, 0, NET_RT_IFLIST, 0 };
    if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0)
        return;
    if (alloc < needed) {
        if (buf != NULL)
            free(buf);
        buf = (char *) malloc(needed);
        if (buf == NULL)
            return;
        alloc = needed;
    }

    if (sysctl(mib, 6, buf, &needed, NULL, 0) < 0)
        return;
    lim = buf + needed;

    next = buf;
    while (next < lim) {
        ifm = (struct if_msghdr *) next;
        if (ifm->ifm_type != RTM_IFINFO)
            return;
        next += ifm->ifm_msglen;

        while (next < lim) {
            nextifm = (struct if_msghdr *) next;
            if (nextifm->ifm_type != RTM_NEWADDR)
                break;
            next += nextifm->ifm_msglen;
        }

        if (ifm->ifm_flags & IFF_UP) {
            sdl = (struct sockaddr_dl *) (ifm + 1);
            if (sdl->sdl_family != AF_LINK)
                continue;
            strncpy(s, sdl->sdl_data, sdl->sdl_nlen);
            s[sdl->sdl_nlen] = 0;

            if (device == s) {
                currIn = ifm->ifm_data.ifi_ibytes;
                currOut = ifm->ifm_data.ifi_obytes;
                break;
            }
        }
    }
#endif
}

// BSD specific code
void KNetProc::readLoadSolaris() {
#ifdef Q_OS_SOLARIS
    static kstat_ctl_t *kc;
    static kstat_t *ksp;
    static kstat_named_t *knp;
    kc = kstat_open();
    ksp = kstat_lookup(kc, 0, -1, device);
    if (ksp && kstat_read(kc, ksp, 0) >= 0) {
        if ((knp = (kstat_named_t *) kstat_data_lookup(ksp, "rbytes")))
            currIn = knp->value.ui32;
        if ((knp = (kstat_named_t *) kstat_data_lookup(ksp, "obytes")))
            currOut = knp->value.ui32;
    }
    kstat_close(kc);
#endif
}

const QStringList &KNetProc::loadInterfaces(bool force)
{
	if ( ! force && ! s_interfaces.empty() )
		return s_interfaces;
	
	if ( force )
		s_interfaces.clear();

#ifndef Q_OS_LINUX
	s_interfaces.push_back("lo");
	s_interfaces.push_back("eth0");
	s_interfaces.push_back("ppp0");
	s_interfaces.push_back("ippp0");
#else
	if ( QDir::root().exists("/sys/class/net") )
	{ // Exists /sys, 2.6 series kernel
		QDir sys("/sys/class/net");
		QStringList l = sys.entryList();
		QStringList::const_iterator end = l.end();
		for(QStringList::const_iterator it = l.begin(); it != end; it++)
		{
			if ( (*it)[0] == '.' ) continue;
			s_interfaces.push_back(*it);
		}
	} else { // Doesn't exists, kernel 2.4 or earlier
		static QFile proc("/proc/net/dev");
		static QRegExp procre("^\\s+(\\S+):(\\d+\\s*){17}$");
		QString read;
		
		if ( ! proc.open(IO_ReadOnly) )
			return s_interfaces;
		
		while ( ! proc.atEnd() )
		{
			if ( proc.readLine(read, MAX_NET_DEV_LINE) == -1 )
			{
				proc.close();
				return s_interfaces;
			}
			
			if ( procre.search(read) == -1 )
				continue;
			
			s_interfaces.push_back(procre.cap(1));
		}
		
		proc.close();
	}
#endif
	return s_interfaces;
}

const QPixmap &KNetProc::getInterfaceIcon(const QString &iface)
{
	QString icon = "devother";
	if ( iface == "lo" )
		icon = "devlo";
	else if ( iface.find("eth") == 0 )
		icon = "deveth0";
	else if ( iface.find("ppp") == 0 )
		icon = "devppp0";
	else if ( iface.find("ippp") == 0 )
		icon = "devippp0";
	
	return KGlobal::iconLoader()->loadIcon(icon, KIcon::Small);
}
